import asyncio
import os
import tempfile

import docker

from apropos.src.bench.bigcodebench.backends.shared import get_imports


async def execute_code_remotely_docker(
    question_dict, code_solution
):  # , packages: List[str] = []
    solution = question_dict["eval_info"]["code_prompt"] + code_solution
    packages, _ = get_imports(solution)
    test = question_dict["eval_info"]["test"]
    complete = solution + "\n" + test

    def get_unit_test_script(path_to_script: str):
        unit_test_script = f"""
import unittest
import io
import os
def test_code():
    path = "{path_to_script}"
    loader = unittest.TestLoader()
    suite = loader.discover('/app', pattern=path)
    runner = unittest.TextTestRunner()
    result = runner.run(suite)

    result_dict = {{
        "errors": len(result.errors),
        "failures": len(result.failures),
        "testsRun": result.testsRun,
        "wasSuccessful": result.wasSuccessful()
    }}

    return result.wasSuccessful(), result_dict
if __name__ == "__main__":
    success, result = test_code()
    print("Success:", success)
    print(result)
"""
        return unit_test_script

    client = docker.from_env()

    with tempfile.TemporaryDirectory() as temp_dir:
        script_path = os.path.join(temp_dir, "script.py")
        test_script_path = os.path.join(temp_dir, "unit_test_script.py")
        with open(script_path, "w") as f:
            f.write(complete)
        with open(test_script_path, "w") as f:
            f.write(get_unit_test_script("script.py"))

        install_command = "apt-get update && apt-get install -y gcc && "
        if packages:
            install_command += f"pip install uv && uv init && uv venv && uv pip install {' '.join(packages)} &&"
        else:
            install_command += "pip install uv && uv init && uv venv &&"
        container = client.containers.run(
            "python:3.9-slim",
            command=f"/bin/bash -c '{install_command} uv run /app/unit_test_script.py'",  # -W ignore
            volumes={temp_dir: {"bind": "/app", "mode": "rw"}},
            working_dir="/app",
            detach=True,
        )

        try:
            container.wait(timeout=60)
            logs = container.logs().decode("utf-8")
        finally:
            container.remove()
    try:
        success = logs.split("Success: ")[1].split("\n")[0] == "True"
        result_dict = eval(logs.split("Success: ")[1].split("\n")[1])
        return success, result_dict
    except Exception as e:
        print("Packages: ", packages)
        print("Please report this error to the developers of apropos-ai @ https://github.com/JoshuaPurtell/Apropos")
        print("Logs: ", logs)
        return False, {
            "errors": 1,
            "failures": 0,
            "testsRun": 0,
            "wasSuccessful": False,
        }


if __name__ == "__main__":
    import asyncio
    from apropos.src.bench.bigcodebench.main import BigCodeBenchComplete_Benchmark

    question = BigCodeBenchComplete_Benchmark().train[0]
    print(
        asyncio.run(
            execute_code_remotely_docker(
                question.information, question.information["answer"], ["pandas"]
            )
        )
    )



import os
import backoff
import together
from together import AsyncTogether, Together
from apropos.src.core.lms.cache_init import cache
from apropos.src.core.lms.vendors.json_structured_outputs.core import (
    add_json_instructions_to_messages,
    extract_pydantic_model_from_response_sync,
    extract_pydantic_model_from_response_async,
)
from apropos.src.core.lms.vendors.openai_like import OpenAIStandardProvider

BACKOFF_TOLERANCE = 1  # 20


class TogetherAPIProvider(OpenAIStandardProvider):
    def __init__(self):
        self.sync_client = Together(api_key=os.getenv("TOGETHER_API_KEY"))
        self.async_client = AsyncTogether(api_key=os.getenv("TOGETHER_API_KEY"))
        self.supports_response_model = False

    @backoff.on_exception(
        backoff.expo,
        (
            together.error.RateLimitError,
            together.error.APIConnectionError,
            together.error.APIError,
        ),
        max_tries=BACKOFF_TOLERANCE,
        giveup=lambda e: getattr(e, "status_code", None) != 429,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        hit = cache.hit_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
        )
        if hit:
            if isinstance(hit, dict) and response_model:
                if "response" in hit:
                    return response_model(**hit["response"])
                return response_model(**hit)
        if hit:
            if isinstance(hit, dict):
                return hit.get("response", hit)
        return hit
        raw_text_api_response = self.sync_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = extract_pydantic_model_from_response_sync(
            raw_text_api_response, response_model
        )
        cache.add_to_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
            structured_api_response,
        )
        return structured_api_response

    @backoff.on_exception(
        backoff.expo,
        (
            together.error.RateLimitError,
            together.error.APIConnectionError,
            together.error.APIError,
        ),
        max_tries=BACKOFF_TOLERANCE,
        giveup=lambda e: getattr(e, "status_code", None) != 429,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        hit = cache.hit_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
        )
        if hit:
            if isinstance(hit, dict) and response_model:
                if "response" in hit:
                    return response_model(**hit["response"])
                return response_model(**hit)
        if hit:
            if isinstance(hit, dict):
                return hit.get("response", hit)
        return hit
        raw_text_api_response = await self.async_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = await extract_pydantic_model_from_response_async(
            raw_text_api_response, response_model
        )
        cache.add_to_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
            structured_api_response,
        )
        return structured_api_response


if __name__ == "__main__":
    import asyncio

    messages = [
        {
            "role": "system",
            "content": "You are a helpful a ssistant that can answer questions abou t the capital of France.",
        },
        {"role": "user", "content": "What  is the capital of France?   "},
    ]
    response = asyncio.run(
        TogetherAPIProvider().async_chat_completion(
            messages=messages,
            model="teknium/OpenHermes-2p5-Mistral-7B",
            temperature=0.0,
            max_tokens=150,
        )
    )
    print(response)



from gym.envs.registration import register

from .bandit import BanditTwoArmedDeterministicFixed
from .bandit import BanditTwoArmedHighHighFixed
from .bandit import BanditTwoArmedHighLowFixed
from .bandit import BanditTwoArmedLowLowFixed

environments = [
    ["BanditTwoArmedDeterministicFixed", "v0"],
    ["BanditTwoArmedHighHighFixed", "v0"],
    ["BanditTwoArmedHighLowFixed", "v0"],
    ["BanditTwoArmedLowLowFixed", "v0"],
]

for environment in environments:
    register(
        id="{}-{}".format(environment[0], environment[1]),
        entry_point="smartplay.bandits:{}".format(environment[0]),
    )



from typing import Dict, Any, Tuple
import tiktoken
from apropos.src.core.lms.cost_table import cost_table_by_model


class CostMonitor:
    token_counts: Dict
    tokenizer: Any
    pricing: Dict

    def __init__(self, model_name):
        self.token_counts = {"system": 0, "user": 0, "response": 0}
        self.tokenizer = (
            tiktoken.encoding_for_model(model_name)
            if "gpt" in model_name
            else tiktoken.encoding_for_model("gpt-4")
        )
        model_name_matches = [k for k in cost_table_by_model.keys() if k in model_name]
        if len(model_name_matches) == 0:
            print(f"Warning: Model {model_name} not found in cost table. Using default pricing of 0.")
            self.pricing = {
                "system": 0,
                "user": 0,
                "response": 0
            }

        else:
            model_name = model_name_matches[0]
        self.pricing = {
            "system": cost_table_by_model[model_name]["cost_per_input_token"],
            "user": cost_table_by_model[model_name]["cost_per_input_token"],
            "response": cost_table_by_model[model_name]["cost_per_output_token"],
        }

    def update_token_counts(self, system_message, user_message, response_message):
        self.token_counts["system"] += len(self.tokenizer.encode(system_message))
        self.token_counts["user"] += len(self.tokenizer.encode(user_message))
        self.token_counts["response"] += len(self.tokenizer.encode(response_message))

    def final_cost(self) -> Tuple[float, int]:
        return sum(
            [self.token_counts[k] * self.pricing[k] for k in self.token_counts]
        ), sum([self.token_counts[k] for k in self.token_counts])



# This file is used to load all games in the smartplay directory.

import warnings
import importlib
import os
import yaml


root_dir = os.path.dirname(os.path.abspath(__file__))
games = []
env_list = []


_exclude_path = ["__pycache__", "utils", "tests", "eval", "example_game"]
_module_dir = os.path.dirname(__file__)


for dirname in os.listdir(root_dir):
    if os.path.isdir(os.path.join(root_dir, dirname)) and dirname not in _exclude_path:
        if "__init__.py" in os.listdir(os.path.join(root_dir, dirname)):
            games.append(dirname)


game_challenges = {}
recorded_settings = {}


for game in games:
    if not os.path.exists(os.path.join(root_dir, game, "evaluation.yml")):
        warnings.warn(
            "Game `{}` does not have evaluation.yml. Skipping the game.".format(game),
            UserWarning,
        )
        continue

    try:
        # Load evaluation settings
        with open(os.path.join(root_dir, game, "evaluation.yml"), "r") as f:
            yml = yaml.safe_load(f)
            recorded_setting = yml["recorded settings"]
            game_challenge = yml["challenges"]

    except Exception as e:
        warnings.warn(
            "Failed to load `evaluation.yml` for `{}`.".format(game), UserWarning
        )
        continue

    try:
        # Load environments
        module = importlib.import_module("." + game, package="smartplay")
        environments = getattr(module, "environments", None)
        if environments:
            for env_name, version in environments:
                env_list.append("{}-{}".format(env_name, version))
                game_challenges[env_name] = (
                    game_challenge["all"]
                    if env_name not in game_challenge.keys()
                    else game_challenge[env_name]
                )
                if env_name in recorded_setting.keys():
                    recorded_settings[env_name] = recorded_setting[env_name]
        else:
            warnings.warn(
                "Failed to load `{}.environments`. Skipping the game.".format(game),
                UserWarning,
            )
            continue

    except Exception as e:
        warnings.warn(
            "Failed to import `{}`. Skipping the game.".format(game), UserWarning
        )
        continue


from .eval import *



from apropos.src.core.lms.helpers import LLM
import numpy as np
from apropos.src.core.utils.program_grounding import (
    Metric,
    messages_to_dag_and_benchmark,
)
from apropos.src.core.optimizers.baselines.bffsrs import BreadthFirstRandomSearch_DAG
import asyncio


def get_hendryks_messages_and_gold_outputs():
    from apropos.src.bench.hendryks_math.main import (
        HendryksMath_Benchmark,
        custom_math_metric,
    )

    benchmark = HendryksMath_Benchmark()
    from apropos.src.bench.hendryks_math.dags.single_step import (
        hendryks_math_single_step_example,
    )

    baseline_single_step_program = hendryks_math_single_step_example(
        model_name="claude-3-haiku-20240307"
    )
    messages_examples = []
    for q in benchmark.train:
        systems_message, user_message = list(
            baseline_single_step_program.nodes.values()
        )[0].transform.prompt.compile(
            inputs={
                "<<<MATHEMATICS_QUESTION>>>": q.information["question"],
            }
        )
        messages_examples.append((systems_message, user_message))
    gold_outputs = [q.information["answer"] for q in benchmark.train]
    return messages_examples, gold_outputs, custom_math_metric


async def optimize_and_score_program(cfg, dag, benchmark, n_to_score: int = 100):
    print("Evaluating baseline program...")
    baseline_program_scores, _ = await benchmark.score_dag(
        dag, n=n_to_score, verbose=True, split="test", patches=["A", "B"]
    )
    print("Baseline program scores:", np.mean(baseline_program_scores))
    bffsrs = BreadthFirstRandomSearch_DAG(
        student_program=dag,
        teacher_program=dag,
        dataset_handler=benchmark,
        cfg=cfg,
    )
    print("Optimizing demonstrations...")
    optimized_dag = await bffsrs.optimize_demonstrations()

    print("Evaluating optimized program...")
    optimized_program_scores, _ = await benchmark.score_dag(
        optimized_dag, n=n_to_score, verbose=True, split="test", patches=["A", "B"]
    )
    print("Optimized program scores:", np.mean(optimized_program_scores))


if __name__ == "__main__":
    # REPLACE WITH YOUR OWN DATA
    # messages_examples: List[List[Dict[str, str]]], len N
    # gold_outputs: List[str], len N
    # metric: Metric
    messages_examples, gold_outputs, custom_math_metric = (
        get_hendryks_messages_and_gold_outputs()
    )

    grounding_llm = LLM(
        "claude-3-5-sonnet-20240620"
    )  # For structuring prompts, better to have high-power
    program_llm = LLM("claude-3-haiku-20240307")  # For generating programs
    metric = Metric(
        gold_outputs_for_dataset=gold_outputs, metric_function=custom_math_metric
    )
    input_names = ["question"]
    name_for_prompt = "Solve Math Problem"
    dag, benchmark = asyncio.run(
        messages_to_dag_and_benchmark(
            messages_examples,
            metric,
            input_names,
            name_for_prompt,
            grounding_llm,
            program_llm,
        )
    )
    bffsrs_config = {
        "optimization": {
            "combination_size": 3,
            "n_iterations": 10,
            "validation_size": 30,
            "test_size": 30,
            "program_search_parallelization_factor": 5,
        },
        "bootstrapping": {
            "patches": ["A", "B"],
            "n_questions": 30,  # 100
        },
        "verbose": True,
    }
    asyncio.run(optimize_and_score_program(bffsrs_config, dag, benchmark))



import time

import backoff
import groq
import instructor
import openai
import together

# from apropos.src.core.lms.api_caching_fxns_diskcache import safecache_wrapper
# from apropos.src.core.lms.cache_init_diskcache import cache, old_cache
# from apropos.src.core.lms.api_caching_fxns_sqlite import #safecache_wrapper
from apropos.src.core.lms.cache_init import cache  # , old_cache
from apropos.src.core.lms.vendors.base import BaseProvider

BACKOFF_TOLERANCE = 200


class OpenAIStandardProvider(BaseProvider):
    def __init__(self, sync_client, async_client):
        self.sync_client = sync_client
        self.async_client = async_client
        self.supports_response_model = False

    @backoff.on_exception(
        backoff.expo,
        (
            openai.RateLimitError,
            together.error.RateLimitError,
            groq.RateLimitError,
            instructor.exceptions.InstructorRetryException,
            together.error.APIConnectionError,
            together.error.APIError,
        ),  # pydantic_core._pydantic_core.ValidationError
        max_tries=BACKOFF_TOLERANCE,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        if not self.supports_response_model:
            raise ValueError(
                "Code for this provider does not yet support response models"
            )
        # hit = safecache_wrapper.hit_cache(messages, model, temperature, response_model, cache, old_cache)
        hit = cache.hit_cache(messages, model, temperature, response_model)
        if hit:
            return hit
        output = self.sync_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_model=response_model,
        )
        cache.add_to_cache(
            messages,
            model,
            temperature,
            response_model,
            output.choices[0].message.content,
        )
        # safecache_wrapper.add_to_cache(messages, model, temperature, response_model, output.choices[0].message.content, cache)
        return output.choices[0].message.content

    @backoff.on_exception(
        backoff.expo,
        (
            openai.RateLimitError,
            together.error.RateLimitError,
            together.error.Timeout,
            groq.RateLimitError,
            instructor.exceptions.InstructorRetryException,
            together.error.APIConnectionError,
            together.error.APIError,
        ),  # pydantic_core._pydantic_core.ValidationError,
        max_tries=BACKOFF_TOLERANCE,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        if not self.supports_response_model:
            raise ValueError(
                "Code for this provider does not yet support response models"
            )
        # hit = safecache_wrapper.hit_cache(messages, model, temperature,  response_model, cache, old_cache)
        hit = cache.hit_cache(messages, model, temperature, response_model)
        # print("Hit true/false: ",hit is not None)
        if hit:
            return hit
        output = await self.async_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_model=response_model,
        )
        # safecache_wrapper.add_to_cache(messages, model, temperature,  response_model, output.choices[0].message.content, cache)
        cache.add_to_cache(
            messages,
            model,
            temperature,
            response_model,
            output.choices[0].message.content,
        )
        return output.choices[0].message.content

    @backoff.on_exception(
        backoff.expo,
        (
            openai.RateLimitError,
            together.error.RateLimitError,
            together.error.Timeout,
            groq.RateLimitError,
            instructor.exceptions.InstructorRetryException,
            together.error.APIConnectionError,
            together.error.APIError,
        ),
        max_tries=BACKOFF_TOLERANCE,
    )
    def sync_chat_completion(self, messages, model, temperature, max_tokens):
        t0 = time.time()
        # hit = safecache_wrapper.hit_cache(messages, model, temperature, None, cache, old_cache)
        hit = cache.hit_cache(messages, model, temperature, None)
        t1 = time.time()
        print(f"Time taken to hit cache: {t1-t0:.2e} seconds")
        if hit:
            return hit
        output = self.sync_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        # safecache_wrapper.add_to_cache(messages, model, temperature, None, output.choices[0].message.content, cache)
        cache.add_to_cache(
            messages, model, temperature, None, output.choices[0].message.content
        )
        return output.choices[0].message.content

    @backoff.on_exception(
        backoff.expo,
        (
            openai.RateLimitError,
            together.error.RateLimitError,
            together.error.Timeout,
            groq.RateLimitError,
            instructor.exceptions.InstructorRetryException,
            together.error.APIConnectionError,
            together.error.APIError,
        ),
        max_tries=BACKOFF_TOLERANCE,
    )
    async def async_chat_completion(self, messages, model, temperature, max_tokens):
        # hit = safecache_wrapper.hit_cache(messages, model, temperature, None, cache, old_cache)
        hit = cache.hit_cache(messages, model, temperature, None)
        if hit:
            return hit
        output = await self.async_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        # safecache_wrapper.add_to_cache(messages, model, temperature, None, output.choices[0].message.content, cache)
        cache.add_to_cache(
            messages, model, temperature, None, output.choices[0].message.content
        )
        return output.choices[0].message.content



import gym
from gym import error, spaces, utils
from gym.utils import seeding
from ..utils import HistoryTracker, describe_act

import random
import itertools
import numpy as np
import messenger
import messenger.envs


""" Format function passed to numpy print to make things pretty.
"""

id_map = {}

for ent in messenger.envs.config.ALL_ENTITIES:
    id_map[ent.id] = ent.name
id_map[0] = "  "
id_map[15] = "you (agent) without the message"
id_map[16] = "you (agent) with the message"


def describe_block(i):
    if i < 17:
        return id_map[i]
    else:
        return "XX"


def describe_loc(ref, P):
    desc = []
    if ref[0] > P[0]:
        desc.append("north")
    elif ref[0] < P[0]:
        desc.append("south")
    if ref[1] > P[1]:
        desc.append("west")
    elif ref[1] < P[1]:
        desc.append("east")

    return "-".join(desc)


def describe_frame(info):
    if 15 in np.unique(info["avatar"]):
        obs = "You (agent) don't have the message."
        agent = 15
    elif 16 in np.unique(info["avatar"]):
        obs = "You (agent) already have the message."
        agent = 16
    else:
        print(np.unique(info["avatar"]))
        raise NotImplemented("Problem with agent")
    center = np.array(np.where(info["avatar"].squeeze() == agent)).squeeze()
    info = info["entities"]
    result = ""
    x = np.arange(info.shape[1])
    y = np.arange(info.shape[0])
    x1, y1 = np.meshgrid(x, y)
    loc = np.stack((y1, x1), axis=-1)
    dist = np.absolute(center - loc).sum(axis=-1)[:, :, np.newaxis]

    obj_info_list = []

    for idx in np.unique(info):
        if idx == 15 or idx == 0:
            continue
        smallest = np.unravel_index(
            np.argmin(np.where(info == idx, dist, np.inf)), info.shape
        )
        obj_info_list.append(
            (
                describe_block(idx),
                dist[smallest[:2]][0],
                describe_loc(np.array([0, 0]), smallest[:2] - center),
            )
        )

    if len(obj_info_list) > 0:
        status_str = "You see:\n{}".format(
            "\n".join(
                [
                    "- {} {} steps to your {}".format(name, dist, loc)
                    if dist > 0
                    else "- {} 0 steps with you".format(name)
                    for name, dist, loc in obj_info_list
                ]
            )
        )
    else:
        status_str = "You see nothing away from you."
    result += obs.strip() + "\n\n" + status_str.strip()

    return result.strip()


class MessengerEnv(gym.Env):
    default_iter = 100
    default_steps = None

    def __init__(self, lvl=1, max_steps=2, env_noise=0):
        lvl_to_steps = [4, 64, 128]
        self.default_steps = lvl_to_steps[lvl - 1]
        env_id = "msgr-test-v{}".format(lvl)
        self._env = gym.make(env_id)
        self.action_list = [
            "Move North",
            "Move South",
            "Move West",
            "Move East",
            "Do Nothing",
        ]

        self.history = HistoryTracker(max_steps)
        self.game_context = """In the game, MESSENGER, each entity can take on one of three roles: an enemy, message, or goal. The agent’s objective is to bring the message to the goal while avoiding the enemies. If the agent encounters an enemy at any point in the game, or the goal without first obtaining the message, it loses the game and obtains a reward of −1.""".strip()
        self.advice = """To solve a game, you may find it helpful to list the objects that you see. Then for each object, match it with an entity description, and identify whether it is good or bad to interact with the object.
The name specifications of in-game objects may not be exact matches. Please try identifying with synonyms.
""".strip()

    def _update_manual(self, manual):
        self.desc = "{}\n\n{}\n\n{}\n\n{}".format(
            self.game_context,
            "\n".join(manual),
            self.advice,
            describe_act(self.action_list),
        ).strip()

    def describe(self, obs, action=None):
        if action is not None:
            return "You took action {}.\n\n{}".format(
                self.action_list[action], describe_frame(obs)
            )
        else:
            return describe_frame(obs)

    def reset(self):
        obs, manual = self._env.reset()
        self._update_manual(manual)
        self.history.reset()
        info = {
            "obs": self.describe(obs),
            "manual": self.desc,
            "history": self.history.describe(),
            "score": 0,
            "completed": 0,
        }
        self.history.step(info)
        return obs, info

    def step(self, action):
        obs, reward, done, info = self._env.step(action)
        try:
            description = self.describe(obs, action)
        except:
            description = "Environment Error."
        info = {
            "obs": description,
            "manual": self.desc,
            "history": self.history.describe(),
            "score": reward,
            "completed": int(reward == 1 and done),
        }
        self.history.step(info)
        return obs, reward, done, info



from openai import OpenAI, AsyncOpenAI
import os
import backoff
import openai
from apropos.src.core.lms.vendors.openai_like import OpenAIStandardProvider
import os
from apropos.src.core.lms.vendors.json_structured_outputs.core import (
    add_json_instructions_to_messages,
    extract_pydantic_model_from_response_sync,
    extract_pydantic_model_from_response_async,
)

BACKOFF_TOLERANCE = 10


class LambdaAPIProvider(OpenAIStandardProvider):
    def __init__(self):
        openai_api_key = os.getenv("LAMBDA_API_KEY")
        openai_api_base = "https://api.lambdalabs.com/v1"
        self.sync_client = OpenAI(
            api_key=openai_api_key,
            base_url=openai_api_base,
        )
        self.async_client = AsyncOpenAI(
            api_key=openai_api_key,
            base_url=openai_api_base,
        )
        self.supports_response_model = True

    @backoff.on_exception(
        backoff.expo,
        (Exception, openai.InternalServerError),
        max_tries=BACKOFF_TOLERANCE,
        giveup=lambda e: getattr(e, "status_code", None) != 429,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        raw_text_api_response = self.sync_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = extract_pydantic_model_from_response_sync(
            raw_text_api_response, response_model
        )
        return structured_api_response

    @backoff.on_exception(
        backoff.expo,
        (Exception, openai.InternalServerError),
        max_tries=BACKOFF_TOLERANCE,
        giveup=lambda e: getattr(e, "status_code", None) != 429,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        raw_text_api_response = await self.async_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = await extract_pydantic_model_from_response_async(
            raw_text_api_response, response_model
        )
        return structured_api_response



import backoff
import openai
import pydantic_core
from openai import AsyncOpenAI, OpenAI
from apropos.src.core.lms.cache_init import cache
from apropos.src.core.lms.vendors.openai_like import OpenAIStandardProvider

import instructor
BACKOFF_TOLERANCE = 100  # 20


class OpenAIAPIProvider(OpenAIStandardProvider):
    def __init__(self, force_structured_output=False):
        self.sync_client = OpenAI()
        self.async_client = AsyncOpenAI()
        self.supports_response_model = True
        #self.force_structured_output = force_structured_output

    @backoff.on_exception(
        backoff.expo,
        (pydantic_core._pydantic_core.ValidationError,),
        max_tries=BACKOFF_TOLERANCE,
        logger=None,
        on_giveup=lambda e: print(e)
        if isinstance(e, pydantic_core._pydantic_core.ValidationError)
        else None,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        hit = cache.hit_cache(messages, model, temperature, None)
        if hit:
            return hit
        sync_client = instructor.patch(OpenAI())
        output = sync_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_model=response_model,
        )
        # HOPEFULLY WE CAN USE THIS INSTEAD OF THE ABOVE
        # output = self.sync_client.beta.chat.completions.parse(
        #     model=model,
        #     messages=messages,
        #     temperature=temperature,
        #     max_tokens=max_tokens,
        #     response_format=response_model,
        # )
        cache.add_to_cache(messages, model, temperature, None, output)
        return output

    @backoff.on_exception(
        backoff.expo,
        (pydantic_core._pydantic_core.ValidationError,),
        max_tries=BACKOFF_TOLERANCE,
        logger=None,
        on_giveup=lambda e: print(e)
        if isinstance(e, pydantic_core._pydantic_core.ValidationError)
        else None,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        hit = cache.hit_cache(messages, model, temperature, None)
        if hit:
            return hit
        async_client = instructor.patch(AsyncOpenAI())
        output = await async_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_model=response_model,
        )
        # output = await self.async_client.beta.chat.completions.parse(
        #     model=model,
        #     messages=messages,
        #     temperature=temperature,
        #     max_tokens=max_tokens,
        #     response_format=response_model,
        # )
        cache.add_to_cache(messages, model, temperature, None, output)
        print("Output:", output)
        return output
from pydantic import BaseModel
from typing import Dict, Any, List

class ActionArgument(BaseModel):
    argument_name: str
    argument_value: Any

class ReAct(BaseModel):
    reasoning: str
    action: str
    action_args: List[ActionArgument]#Dict[str, Any]

async def react_example():
    messages = [
        {
            "role": "system",
            "content": "You are an AI assistant tasked with planning a trip. Use the ReAct pattern to break down your thinking and actions.",
        },
        {"role": "user", "content": "Plan a 3-day trip to Paris."},
    ]
    
    openai_provider = OpenAIAPIProvider()
    response = await openai_provider.async_chat_completion_with_response_model(
        messages=messages,
        model="gpt-4o-2024-08-06",
        temperature=0.7,
        max_tokens=300,
        response_model=ReAct
    )
    
    print(f"Reasoning: {response.reasoning}")
    print(f"Action: {response.action}")
    print(f"Action Args: {response.action_args}")


if __name__ == "__main__":
    import asyncio

    messages = [
        {
            "role": "system",
            "content": "You are a helpful assistant that can answer questions about the capital of France.",
        },
        {"role": "user", "content": "What is the capital of France?"},
    ]
    response = asyncio.run(
        OpenAIAPIProvider().async_chat_completion(
            messages=messages,
            model="gpt-4o-mini",
            temperature=0.0,
            max_tokens=500,
        )
    )
    print(response)
    asyncio.run(react_example())



action_dict = {
    i: a
    for i, a in enumerate(
        [
            "Noop",
            "Move West",
            "Move East",
            "Move North",
            "Move South",
            "Do",
            "Sleep",
            "Place Stone",
            "Place Table",
            "Place Furnace",
            "Place Plant",
            "Make Wood Pickaxe",
            "Make Stone Pickaxe",
            "Make Iron Pickaxe",
            "Make Wood Sword",
            "Make Stone Sword",
            "Make Iron Sword",
        ]
    )
}

crafter_game_rules = {
    "World Information": {
        "Actions with Requirements": {
            "Movement": {
                "Move West": {"Requirements": "Flat ground west of the agent."},
                "Move East": {"Requirements": "Flat ground east of the agent."},
                "Move North": {"Requirements": "Flat ground north of the agent."},
                "Move South": {"Requirements": "Flat ground south of the agent."},
            },
            "Basic Actions": {
                "Do": {
                    "Requirements": "Facing creature or material; have necessary tool."
                },
                "Sleep": {"Requirements": "Energy level is below maximum."},
                "Noop": {"Requirements": "Always applicable. (Means no operation.)"},
            },
            "Crafting and Building": {
                "Place Stone": {"Requirements": "Stone in inventory."},
                "Place Table": {"Requirements": "Wood in inventory."},
                "Place Furnace": {"Requirements": "Stone in inventory."},
                "Place Plant": {"Requirements": "Sapling in inventory."},
                "Make Wood Pickaxe": {
                    "Requirements": "Nearby table; wood in inventory."
                },
                "Make Stone Pickaxe": {
                    "Requirements": "Nearby table; wood, stone in inventory."
                },
                "Make Iron Pickaxe": {
                    "Requirements": "Nearby table, furnace; wood, coal, iron in inventory."
                },
                "Make Wood Sword": {"Requirements": "Nearby table; wood in inventory."},
                "Make Stone Sword": {
                    "Requirements": "Nearby table; wood, stone in inventory."
                },
                "Make Iron Sword": {
                    "Requirements": "Nearby table, furnace; wood, coal, iron in inventory."
                },
            },
        },
        "Hierarchy of Goals": {
            "Collect Wood": {
                "Place Table": {
                    "Make Wood Pickaxe": {
                        "Collect Stone": {
                            "Place Stone": {},
                            "Make Stone Pickaxe": {
                                "Collect Iron": {
                                    "Make Iron Pickaxe": {
                                        "Collect Diamond": {},
                                        "Place Furnace": {
                                            "Make Iron Sword": {},
                                            "Make Iron Pickaxe": {},
                                        },
                                    }
                                },
                                "Collect Coal": {},
                            },
                        }
                    },
                    "Make Wood Sword": {},
                }
            },
            "Eat Cow": {},
            "Collect Sapling": {"Place Plant": {"Eat Plant": {}}},
            "Collect Drink": {},
            "Defeat Zombie": {},
            "Defeat Skeleton": {},
            "Wake Up": {},
        },
    }
}

crafter_game_tips = [
    "Don't take actions for which you don't have the necessary requirements. For example, don't try to place a table if you don't have wood in your inventory.",
    "Place safety and survival first, but when the coast is clear strive to make progress on unlocking the next goals in your progression.",
]



from pydantic import BaseModel
from apropos.src.core.utils.program_grounding import build_single_step_program
from apropos.src.core.programs.prompt import (
    PromptTemplate,
    SystemMessage,
    UserMessage,
    Topic,
)
from apropos.src.core.programs.dag import LM_DAG
import asyncio


class StepQualityScorer(BaseModel):
    score: int
    reasoning: str


def get_single_step_prm_dag_for_math_tree_search(
    model_name: str,
) -> LM_DAG:
    prompt = PromptTemplate(
        name="step_quality_scorer",
        system=SystemMessage(
            premise=[
                Topic(
                    topic_name="premise",
                    topic_template="You will be given a $MATH_QUESTION, a $COT, and a $CANDIDATE_STEP.",
                    instructions_fields={
                        "$MATH_QUESTION": "mathematics question",
                        "$COT": "chain of atomic reasoning steps",
                        "$CANDIDATE_STEP": "proposed next reasoning step",
                    },
                    input_fields=[],
                )
            ],
            objective=[
                Topic(
                    topic_name="objective",
                    topic_template="Your goal is to evaluate the quality of the $CANDIDATE_STEP and assign it a score from 0 to 10.",
                    instructions_fields={
                        "$CANDIDATE_STEP": "proposed next reasoning step"
                    },
                    input_fields=[],
                )
            ],
            constraints=[
                Topic(
                    topic_name="constraints",
                    topic_template="Consider the following criteria:\n- $RELEVANCE\n- $CORRECTNESS\n- $CLARITY\n- $PROGRESS\nProvide only the numerical score as output.",
                    instructions_fields={
                        "$RELEVANCE": "How relevant is the step to solving the problem?",
                        "$CORRECTNESS": "Is the step mathematically correct?",
                        "$CLARITY": "Is the step clear and easy to understand?",
                        "$PROGRESS": "Does the step make meaningful progress towards the solution?",
                    },
                    input_fields=[],
                )
            ],
        ),
        user=UserMessage(
            user=[
                Topic(
                    topic_name="user_input",
                    topic_template="$MATH_QUESTION: <<<MATHEMATICS_QUESTION>>>\n$COT:\n<<<CHAIN_OF_THOUGHT>>>\n$CANDIDATE_STEP: <<<CANDIDATE_STEP>>>\n\nQuality score (0-10):",
                    instructions_fields={},
                    input_fields=[
                        "<<<MATHEMATICS_QUESTION>>>",
                        "<<<CHAIN_OF_THOUGHT>>>",
                        "<<<CANDIDATE_STEP>>>",
                    ],
                )
            ],
        ),
        response_type="pydantic",
        response_model_scheme=StepQualityScorer.schema(),
        demonstrations=[],
    )

    single_step_prm_program = build_single_step_program(
        model_name=model_name,
        prompt=prompt,
        dag_input_names=[
            "<<<MATHEMATICS_QUESTION>>>",
            "<<<CHAIN_OF_THOUGHT>>>",
            "<<<CANDIDATE_STEP>>>",
        ],
        dag_input_aliases={
            "question": "<<<MATHEMATICS_QUESTION>>>",
            "cot": "<<<CHAIN_OF_THOUGHT>>>",
            "candidate_step": "<<<CANDIDATE_STEP>>>",
        },
        dag_output_aliases={"<<<FINAL_ANSWER>>>": "score"},
    )
    return single_step_prm_program


if __name__ == "__main__":
    model_name = "gpt-4o"
    single_step_prm_program = get_single_step_prm_dag_for_math_tree_search(model_name)
    inputs = {
        "<<<MATHEMATICS_QUESTION>>>": "What is the sum of 1 and 1?",
        "<<<CHAIN_OF_THOUGHT>>>": "1 + 1 = 2",
        "<<<CANDIDATE_STEP>>>": "2",
    }
    print(asyncio.run(single_step_prm_program.arun(inputs)))



import numpy as np
import gym
from gym import spaces
from gym.utils import seeding
from ..utils import HistoryTracker, describe_act
import random


class BanditEnv(gym.Env):
    """
    Bandit environment base to allow agents to interact with the class n-armed bandit
    in different variations

    p_dist:
        A list of probabilities of the likelihood that a particular bandit will pay out
    r_dist:
        A list of either rewards (if number) or means and standard deviations (if list)
        of the payout that bandit has
    """

    default_iter = 20
    default_steps = 50

    def __init__(self, p_dist, r_dist, max_steps=50):
        if len(p_dist) != len(r_dist):
            raise ValueError(
                "Probability and Reward distribution must be the same length"
            )

        if min(p_dist) < 0 or max(p_dist) > 1:
            raise ValueError("All probabilities must be between 0 and 1")

        for reward in r_dist:
            if isinstance(reward, list) and reward[1] <= 0:
                raise ValueError(
                    "Standard deviation in rewards must all be greater than 0"
                )

        self.n_bandits = len(p_dist)
        self.action_space = spaces.Discrete(self.n_bandits)
        self.action_list = [
            "Pull slot machine {}.".format(i + 1) for i in range(self.n_bandits)
        ]
        self.observation_space = spaces.Discrete(1)
        self.p_dist = p_dist
        self.r_dist = r_dist
        self.desc = """
You are in the casino with 2 slot machines in front of you. Your goal is to try to earn the most from those slot machines.

{}
""".format(describe_act(self.action_list)).strip()
        self.history = HistoryTracker(max_steps)

        self._seed()
        self.score_tracker = 0

    def _seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        assert self.action_space.contains(action)

        reward = -1
        done = False

        if np.random.uniform() < self.p_dist[action]:
            if not isinstance(self.r_dist[action], list):
                reward = self.r_dist[action]
            else:
                reward = np.random.normal(
                    self.r_dist[action][0], self.r_dist[action][1]
                )

        self.score_tracker += int(action == self.optimal)
        info = {
            "obs": "You pulled slot machine {}, you received reward {}.".format(
                action + 1, reward
            ),
            "score": self.score_tracker,
            "manual": self.desc,
            "history": self.history.describe(),
            "optimal": self.optimal,
            "completed": 0,
        }
        self.history.step(info)

        return 0, reward, done, info

    def reset(self):
        idx_list = list(range(len(self.p_dist)))
        random.shuffle(idx_list)

        self.score_tracker = 0
        self.p_dist = [self.p_dist[i] for i in idx_list]
        self.r_dist = [self.r_dist[i] for i in idx_list]
        self.ev = [p * r - (1 - p) for p, r in zip(self.p_dist, self.r_dist)]
        self.optimal = np.argmax(self.ev)
        self.history.reset()
        info = {
            "obs": "A new round begins.",
            "score": self.score_tracker,
            "manual": self.desc,
            "history": self.history.describe(),
            "optimal": self.optimal,
            "completed": 0,
        }
        self.history.step(info)
        return 0, info

    def render(self, mode="human", close=False):
        pass


class BanditTwoArmedDeterministicFixed(BanditEnv):
    """Simplest case where one bandit always pays, and the other always doesn't"""

    def __init__(self):
        BanditEnv.__init__(self, p_dist=[1, 0], r_dist=[1, 1])


class BanditTwoArmedHighLowFixed(BanditEnv):
    """Stochastic version with a large difference between which bandit pays out of two choices"""

    def __init__(self):
        BanditEnv.__init__(self, p_dist=[0.8, 0.2], r_dist=[1, 1])


class BanditTwoArmedHighHighFixed(BanditEnv):
    """Stochastic version with a small difference between which bandit pays where both are good"""

    def __init__(self):
        BanditEnv.__init__(self, p_dist=[0.8, 0.9], r_dist=[1, 1])


class BanditTwoArmedLowLowFixed(BanditEnv):
    """Stochastic version with a small difference between which bandit pays where both are bad"""

    def __init__(self):
        BanditEnv.__init__(self, p_dist=[0.1, 0.2], r_dist=[1, 1])



import datetime
import json
import pathlib

import imageio
import numpy as np


class Recorder:
    def __init__(
        self,
        env,
        directory,
        save_stats=True,
        save_video=True,
        save_episode=True,
        video_size=(512, 512),
    ):
        if directory and save_stats:
            env = StatsRecorder(env, directory)
        if directory and save_video:
            env = VideoRecorder(env, directory, video_size)
        if directory and save_episode:
            env = EpisodeRecorder(env, directory)
        self._env = env

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        return getattr(self._env, name)


class StatsRecorder:
    def __init__(self, env, directory):
        self._env = env
        self._directory = pathlib.Path(directory).expanduser()
        self._directory.mkdir(exist_ok=True, parents=True)
        self._file = (self._directory / "stats.jsonl").open("a")
        self._length = None
        self._reward = None
        self._unlocked = None
        self._stats = None

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        return getattr(self._env, name)

    def reset(self):
        obs = self._env.reset()
        self._length = 0
        self._reward = 0
        self._unlocked = None
        self._stats = None
        return obs

    def step(self, action):
        obs, reward, done, info = self._env.step(action)
        self._length += 1
        self._reward += info["reward"]
        if done:
            self._stats = {"length": self._length, "reward": round(self._reward, 1)}
            for key, value in info["achievements"].items():
                self._stats[f"achievement_{key}"] = value
            self._save()
        return obs, reward, done, info

    def _save(self):
        self._file.write(json.dumps(self._stats) + "\n")
        self._file.flush()


class VideoRecorder:
    def __init__(self, env, directory, size=(512, 512)):
        if not hasattr(env, "episode_name"):
            env = EpisodeName(env)
        self._env = env
        self._directory = pathlib.Path(directory).expanduser()
        self._directory.mkdir(exist_ok=True, parents=True)
        self._size = size
        self._frames = None

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        return getattr(self._env, name)

    def reset(self):
        obs = self._env.reset()
        self._frames = [self._env.render(self._size)]
        return obs

    def step(self, action):
        obs, reward, done, info = self._env.step(action)
        self._frames.append(self._env.render(self._size))
        if done:
            self._save()
        return obs, reward, done, info

    def _save(self):
        filename = str(self._directory / (self._env.episode_name + ".mp4"))
        imageio.mimsave(filename, self._frames)


class EpisodeRecorder:
    def __init__(self, env, directory):
        if not hasattr(env, "episode_name"):
            env = EpisodeName(env)
        self._env = env
        self._directory = pathlib.Path(directory).expanduser()
        self._directory.mkdir(exist_ok=True, parents=True)
        self._episode = None

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        return getattr(self._env, name)

    def reset(self):
        obs = self._env.reset()
        self._episode = [{"image": obs}]
        return obs

    def step(self, action):
        # Transitions are defined from the environment perspective, meaning that a
        # transition contains the action and the resulting reward and next
        # observation produced by the environment in response to said action.
        obs, reward, done, info = self._env.step(action)
        transition = {
            "action": action,
            "image": obs,
            "reward": reward,
            "done": done,
        }
        for key, value in info.items():
            if key in ("inventory", "achievements"):
                continue
            transition[key] = value
        for key, value in info["achievements"].items():
            transition[f"achievement_{key}"] = value
        for key, value in info["inventory"].items():
            transition[f"ainventory_{key}"] = value
        self._episode.append(transition)
        if done:
            self._save()
        return obs, reward, done, info

    def _save(self):
        filename = str(self._directory / (self._env.episode_name + ".npz"))
        # Fill in zeros for keys missing at the first time step.
        for key, value in self._episode[1].items():
            if key not in self._episode[0]:
                self._episode[0][key] = np.zeros_like(value)
        episode = {
            k: np.array([step[k] for step in self._episode]) for k in self._episode[0]
        }
        np.savez_compressed(filename, **episode)


class EpisodeName:
    def __init__(self, env):
        self._env = env
        self._timestamp = None
        self._unlocked = None
        self._length = None

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        return getattr(self._env, name)

    def reset(self):
        obs = self._env.reset()
        self._timestamp = None
        self._unlocked = None
        self._length = 0
        return obs

    def step(self, action):
        obs, reward, done, info = self._env.step(action)
        self._length += 1
        if done:
            self._timestamp = datetime.datetime.now().strftime("%Y%m%dT%H%M%S")
            self._unlocked = sum(int(v >= 1) for v in info["achievements"].values())
        return obs, reward, done, info

    @property
    def episode_name(self):
        return f"{self._timestamp}-ach{self._unlocked}-len{self._length}"



from gym.envs.registration import register

from .crafter_env import Crafter

environments = [
    ["Crafter", "v0"],
]

for environment in environments:
    register(
        id="{}-{}".format(environment[0], environment[1]),
        entry_point="smartplay.crafter:{}".format(environment[0]),
        kwargs={"reward": True},
    )
